from typing import Optional

from web3.types import TxParams
from web3.contract.contract import ContractFunction
from web3 import Web3
from eth_typing import HexStr
from eth_account import Account

from .types import BatchOperationType


class Operation:

    def __init__(self, rpc: str, agreement_call: ContractFunction, type: BatchOperationType, forwarder_call: Optional[ContractFunction] = None) -> None:
        self.rpc = rpc
        self.agreement_call = agreement_call
        self.type = type
        self.forwarder_call = forwarder_call

    def exec(self, private_key: str) -> HexStr:
        """
            Signs and broadcasts a transaction
            @param rpc - rpc url
            @param private_key - private key
            @returns - HexStr - The transaction hash
        """
        populated_transaction = self._get_populated_transaction_request(
            self.rpc, private_key)
        web3 = Web3(Web3.HTTPProvider(self.rpc))
        signed_txn = web3.eth.account.sign_transaction(
            populated_transaction, private_key=private_key)
        transaction_hash = web3.eth.send_raw_transaction(
            signed_txn.rawTransaction)
        return transaction_hash.hex()

    def _get_populated_transaction_request(self, rpc: str, private_key: str) -> TxParams:
        """
            Selects and prepares the transaction object to be signed
            @param rpc - rpc url
            @param private_key - private key
            @returns - TxParams - The transaction object
        """
        call = self.forwarder_call if self.forwarder_call is not None else self.agreement_call
        address = Account.from_key(private_key).address
        populated_transaction = call.build_transaction({
            "from": address
        })
        web3 = Web3(Web3.HTTPProvider(rpc))
        nonce = web3.eth.get_transaction_count(address)
        populated_transaction["nonce"] = nonce
        return populated_transaction



from eth_typing import HexAddress
from web3 import Web3

from .errors import InvalidAddressError, InvalidChainId
from .constants import NETWORKS, AUTHORIZE_FLOW_OPERATOR_CREATE, AUTHORIZE_FLOW_OPERATOR_DELETE, AUTHORIZE_FLOW_OPERATOR_UPDATE
# from .types import NETWORK


def to_bytes32(string: str) -> bytes:
    encoded_string = string.encode('utf-8')
    if len(encoded_string) > 32:
        raise ValueError("Input string is too long for bytes32")
    padding_length = 32 - len(encoded_string)
    padded_string = encoded_string + b'\x00' * padding_length
    return padded_string


def to_bytes(string: str) -> bytes:
    encoded_string = string.encode('utf-8')
    return encoded_string


def normalize_address(address: HexAddress) -> HexAddress:
    if len(address) == 42 or 40:
        return Web3.to_checksum_address(address)
    else:
        raise InvalidAddressError(f"{address} is invalid")


def is_permissions_clean(permissions: int) -> bool:
    return ((permissions & ~(AUTHORIZE_FLOW_OPERATOR_CREATE | AUTHORIZE_FLOW_OPERATOR_UPDATE | AUTHORIZE_FLOW_OPERATOR_DELETE)) == 0)


def validate_chain_id(chain_id: int) -> bool:
    for network in NETWORKS:
        if chain_id == network.get("chainId"):
            return True
    raise InvalidChainId("Chain not supported")


def get_network(chain_id: int):
    from .types import NETWORK
    for network in NETWORKS:
        if chain_id == network.get("chainId"):
            return NETWORK(network)
    raise InvalidChainId("Chain not supported")



import unittest

from main.superfluid import is_permissions_clean, validate_chain_id, get_network
from main.superfluid import InvalidChainId
from main.superfluid import types


class TestPermissionChecker(unittest.TestCase):

    def test_valid_permissions(self):
        self.assertEqual(is_permissions_clean(
            1), True)
        self.assertEqual(is_permissions_clean(
            2), True)
        self.assertEqual(is_permissions_clean(
            3), True)
        self.assertEqual(is_permissions_clean(
            4), True)
        self.assertEqual(is_permissions_clean(
            5), True)
        self.assertEqual(is_permissions_clean(
            6), True)
        self.assertEqual(is_permissions_clean(
            7), True)

    def test_invalid_permissions(self):
        self.assertEqual(is_permissions_clean(
            8), False)
        self.assertEqual(is_permissions_clean(
            9), False)
        self.assertEqual(is_permissions_clean(
            10), False)
        self.assertEqual(is_permissions_clean(
            11), False)
        self.assertEqual(is_permissions_clean(
            12), False)
        self.assertEqual(is_permissions_clean(
            13), False)
        self.assertEqual(is_permissions_clean(
            14), False)


class TestValidateChainId(unittest.TestCase):

    def test_valid_chain_id(self):
        self.assertEqual(validate_chain_id(
            5), True)
        self.assertEqual(validate_chain_id(
            80001), True)
        self.assertEqual(validate_chain_id(
            420), True)
        self.assertEqual(validate_chain_id(
            421613), True)
        self.assertEqual(validate_chain_id(
            43113), True)
        self.assertEqual(validate_chain_id(
            11155111), True)
        self.assertEqual(validate_chain_id(
            100), True)

    def test_invalid_chain_id(self):
        with self.assertRaises(InvalidChainId):
            validate_chain_id(88)
        with self.assertRaises(InvalidChainId):
            validate_chain_id(599)
        with self.assertRaises(InvalidChainId):
            validate_chain_id(76)


class TestGetNetwork(unittest.TestCase):

    def test_valid_network(self):
        polygon_mumbai_network = get_network(80001)
        self.assertEqual(
            polygon_mumbai_network.RESOLVER, "0x8C54C83FbDe3C59e59dd6E324531FB93d4F504d3")
        self.assertEqual(
            polygon_mumbai_network.HOST, "0xEB796bdb90fFA0f28255275e16936D25d3418603")
        self.assertEqual(
            polygon_mumbai_network.GOVERNANCE, "0x2637eA93EE5cd887ff9AC98185eA67Bd70C5f62e")
        self.assertEqual(
            polygon_mumbai_network.CFA_V1, "0x49e565Ed1bdc17F3d220f72DF0857C26FA83F873")
        self.assertEqual(
            polygon_mumbai_network.CFA_V1_FORWARDER, "0xcfA132E353cB4E398080B9700609bb008eceB125")
        self.assertEqual(
            polygon_mumbai_network.IDA_V1, "0x804348D4960a61f2d5F9ce9103027A3E849E09b8")
        self.assertEqual(
            polygon_mumbai_network.SUPER_TOKEN_FACTORY, "0xB798553db6EB3D3C56912378409370145E97324B")
        self.assertEqual(
            polygon_mumbai_network.SUPERFLUID_LOADER, "0x0d56ED56b63382B0FC964490feB9AE438B6B4b79")
        self.assertEqual(
            polygon_mumbai_network.TOGA, "0x38DD80876DBA048d0050D28828522c313967D073")
        self.assertEqual(
            polygon_mumbai_network.SUPER_SPREADER, "0x74CDF863b00789c29734F8dFd9F83423Bc55E4cE")
        self.assertEqual(
            polygon_mumbai_network.FLOW_SCHEDULER, "0x59A3Ba9d34c387FB70b4f4e4Fbc9eD7519194139")
        self.assertEqual(
            polygon_mumbai_network.VESTING_SCHEDULER, "0x3962EE56c9f7176215D149938BA685F91aBB633B")

        optimism_goerli_network = get_network(420)
        self.assertEqual(
            optimism_goerli_network.RESOLVER, "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136")
        self.assertEqual(
            optimism_goerli_network.HOST, "0xE40983C2476032A0915600b9472B3141aA5B5Ba9")
        self.assertEqual(
            optimism_goerli_network.GOVERNANCE, "0x777Be25F9fdcA87e8a0E06Ad4be93d65429FCb9f")
        self.assertEqual(
            optimism_goerli_network.CFA_V1, "0xff48668fa670A85e55A7a822b352d5ccF3E7b18C")
        self.assertEqual(
            optimism_goerli_network.CFA_V1_FORWARDER, "0xcfA132E353cB4E398080B9700609bb008eceB125")
        self.assertEqual(
            optimism_goerli_network.IDA_V1, "0x96215257F2FcbB00135578f766c0449d239bd92F")
        self.assertEqual(
            optimism_goerli_network.SUPER_TOKEN_FACTORY, "0xfafe31cf998Df4e5D8310B03EBa8fb5bF327Eaf5")
        self.assertEqual(
            optimism_goerli_network.SUPERFLUID_LOADER, "0x5Bb5908dcCC9Bb0fC39a78CfDf9e47B4C08E9521")
        self.assertEqual(
            optimism_goerli_network.TOGA, None)
        self.assertEqual(
            optimism_goerli_network.SUPER_SPREADER, None)
        self.assertEqual(
            optimism_goerli_network.FLOW_SCHEDULER, None)
        self.assertEqual(
            optimism_goerli_network.VESTING_SCHEDULER, None)

    def test_invalid_network(self):
        with self.assertRaises(InvalidChainId):
            get_network(59)
        with self.assertRaises(InvalidChainId):
            get_network(500)
        with self.assertRaises(InvalidChainId):
            get_network(60)



import json
import os

from .abis import cfa_v1, cfa_v1_forwarder, host
from .metadata import networks


CFA_V1_ABI = json.loads(cfa_v1)

CFA_V1_FORWARDER_ABI = json.loads(cfa_v1_forwarder)

HOST_ABI = json.loads(host)

NETWORKS = json.loads(networks)

"""
------- ACL AUTHORIZATION BIT OPERATIONS -------
"""
AUTHORIZE_FLOW_OPERATOR_CREATE = 1 << 0
AUTHORIZE_FLOW_OPERATOR_UPDATE = 1 << 1
AUTHORIZE_FLOW_OPERATOR_DELETE = 1 << 2



from typing import Optional

from web3 import Web3
from web3.types import TxParams
from web3.middleware import geth_poa_middleware
from web3.contract.contract import ContractFunction
from eth_typing import HexAddress, HexStr

from .host import Host
from .constants import CFA_V1_ABI, CFA_V1_FORWARDER_ABI
from .types import GetFlowParams, GetAccountFlowInfoParams, GetFlowOperatorDataParams, GetFlowOperatorDataParamsByID, CreateFlowParams, UpdateFlowParams, DeleteFlowParams, CreateFlowByOperatorParams, UpdateFlowByOperatorParams, Web3FlowInfo, UpdateFlowParams, Web3FlowOperatorData, FlowRateAllowanceParams, UpdateFlowOperatorPermissionsParams, FullControlParams
from .errors import SFError
from .operation import Operation
from .utils import get_network


class CFA_V1:

    def __init__(self, rpc: str, chain_id: int) -> None:
        self.rpc = rpc
        network = get_network(chain_id)
        web3 = Web3(Web3.HTTPProvider(rpc))
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.host = Host(rpc, network.HOST)
        self.contract = web3.eth.contract(
            address=network.CFA_V1, abi=CFA_V1_ABI)
        self.forwarder = web3.eth.contract(
            address=network.CFA_V1_FORWARDER, abi=CFA_V1_FORWARDER_ABI)

    def get_flow(self, super_token: HexAddress, sender: HexAddress, receiver: HexAddress) -> Web3FlowInfo:
        """
            Get the details of a flow.

            @param super_token - The token to be flowed
            @param sender - the sender of the flow
            @param receiver - the receiver of the flow
            @returns - Web3FlowInfo
        """
        try:
            validated_params = GetFlowParams(super_token, sender, receiver)
            transaction_response = self.contract.functions.getFlow(
                validated_params.super_token, validated_params.sender, validated_params.receiver).call()
            info = {
                "timestamp": transaction_response[0],
                "flowRate": transaction_response[1],
                "deposit": transaction_response[2],
                "owedDeposit": transaction_response[3]
            }
            return info
        except Exception as e:
            raise SFError(
                "CFAV1_READ", "There was an error getting the flow", e)

    def get_account_flow_info(self, super_token: HexAddress, account: HexAddress) -> Web3FlowInfo:
        """
            Get the details of a account flow in a super token

            @param super_token - The token to be flowed
            @param account - the account to get its info
            @returns - Web3FlowInfo
        """
        try:
            validated_params = GetAccountFlowInfoParams(
                super_token, account)
            transaction_response = self.contract.functions.getAccountFlowInfo(
                validated_params.super_token, validated_params.account).call()
            info = {
                "timestamp": transaction_response[0],
                "flowRate": transaction_response[1],
                "deposit": transaction_response[2],
                "owedDeposit": transaction_response[3]
            }
            return info
        except Exception as e:
            raise SFError(
                "CFAV1_READ", "There was an error getting the account flow information", e)

    def get_net_flow(self, super_token: HexAddress, account: HexAddress) -> int:
        """
            Get the details of the net flow of an account in a super token.

            @param super_token - The token to be flowed
            @param account - the account to get its info
            @returns - int: net flow rate of the account
        """
        try:
            validated_params = GetAccountFlowInfoParams(
                super_token, account)
            transaction_response = self.contract.functions.getNetFlow(
                validated_params.super_token, validated_params.account).call()
            net_flow_rate = transaction_response
            return net_flow_rate
        except Exception as e:
            raise SFError(
                "CFAV1_READ", "There was an error getting net flow", e)

    def get_flow_operator_data(self, super_token: HexAddress, sender: HexAddress, flow_operator: HexAddress) -> Web3FlowOperatorData:
        """
            Get the details of a flow operator to a sender

            @param super_token - The token to be flowed
            @param sender - the sender of the flow
            @param flow_operator - the spender
            @returns - Web3FlowOperatorData
        """
        try:
            validated_params = GetFlowOperatorDataParams(
                super_token, sender, flow_operator)
            transaction_response = self.contract.functions.getFlowOperatorData(
                validated_params.super_token, validated_params.sender, validated_params.flow_operator).call()
            flow_operator_data = {
                # TODO: Review conversions
                "flowOperatorId": Web3.to_hex(transaction_response[0]),
                "permissions": transaction_response[1],
                "flowRateAllowance": transaction_response[2]
            }
            return flow_operator_data
        except Exception as e:
            raise SFError(
                "CFAV1_READ", "There was an error getting flow operator data", e)

    def get_flow_operator_data_by_id(self, super_token: HexAddress, flow_operator_id: HexStr) -> Web3FlowOperatorData:
        """
            Get the details of a flow operator to a sender by id

            @param super_token - The token to be flowed
            @param flow_operator_id - the flow operator id
            @returns - Web3FlowOperatorData
        """
        try:
            validated_params = GetFlowOperatorDataParamsByID(
                super_token, flow_operator_id)
            transaction_response = self.contract.functions.getFlowOperatorDataByID(
                validated_params.super_token, validated_params.flow_operator_id).call()
            flow_operator_data = {
                # TODO: Review conversions
                "flowOperatorId": validated_params.flow_operator_id,
                "permissions": transaction_response[0],
                "flowRateAllowance": transaction_response[1]
            }
            return flow_operator_data
        except Exception as e:
            raise SFError(
                "CFAV1_READ", "There was an error getting flow operator data", e)

    def create_flow(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, flow_rate: int, user_data: Optional[HexStr] = None,  should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Creates a flow

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param flow_rate - flow rate for the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = CreateFlowParams(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='createFlow', args=[
            validated_params.super_token, validated_params.receiver, validated_params.flow_rate, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.createFlow(
            validated_params.super_token, validated_params.sender, validated_params.receiver, validated_params.flow_rate, validated_params.user_data or "0x")
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def update_flow(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, flow_rate: int, user_data: Optional[HexStr] = None,  should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Updates a flow

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param flow_rate - flow rate for the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = UpdateFlowParams(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='updateFlow', args=[
            validated_params.super_token, validated_params.receiver, validated_params.flow_rate, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.updateFlow(
            validated_params.super_token, validated_params.sender, validated_params.receiver, validated_params.flow_rate, validated_params.user_data or "0x")
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def delete_flow(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Deletes a flow

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = DeleteFlowParams(
            sender, receiver, super_token, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='deleteFlow', args=[
            validated_params.super_token, validated_params.sender, validated_params.receiver, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.deleteFlow(
            validated_params.super_token, validated_params.sender, validated_params.receiver, validated_params.user_data or "0x")
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def increase_flow_rate_allowance(self, super_token: HexAddress, flow_operator: HexAddress, flow_rate_allowance_delta: int, user_data: Optional[HexStr] = None) -> Operation:
        """
            Increases the flow rate allowance of a flow operator

            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param flow_rate_allowance_delta - The amount to increase the flow rate allowance by
            @param user_data(Optional) - Extra data provided
            @returns - Operation
        """
        validated_params = FlowRateAllowanceParams(
            super_token, flow_operator, flow_rate_allowance_delta, user_data)
        calldata = self.contract.encodeABI(fn_name='increaseFlowRateAllowance', args=[
            validated_params.super_token, validated_params.flow_operator, validated_params.flow_rate_allowance_delta, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        return call_agreement_operation

    def decrease_flow_rate_allowance(self, super_token: HexAddress, flow_operator: HexAddress, flow_rate_allowance_delta: int, user_data: Optional[HexStr] = None) -> Operation:
        """
            Decreases the flow rate allowance of a flow operator

            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param flow_rate_allowance_delta - The amount to decrease the flow rate allowance by
            @param user_data(Optional) - Extra data provided
            @returns - Operation
        """
        validated_params = FlowRateAllowanceParams(
            super_token, flow_operator, flow_rate_allowance_delta, user_data)
        calldata = self.contract.encodeABI(fn_name='decreaseFlowRateAllowance', args=[
            validated_params.super_token, validated_params.flow_operator, validated_params.flow_rate_allowance_delta, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        return call_agreement_operation

    def update_flow_operator_permissions(self, super_token: HexAddress, flow_operator: HexAddress, permissions: int, flow_rate_allowance: int, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Update permissions for a flow operator as a sender.

            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param permissions - Number specifying the permission type
            @param flow_rate_allowance - Allowance to an operator
            @param user_data(Optional) - Extra data provided
            @returns - Operation
        """
        validated_params = UpdateFlowOperatorPermissionsParams(
            super_token, flow_operator, permissions, flow_rate_allowance, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='updateFlowOperatorPermissions', args=[
            validated_params.super_token, validated_params.flow_operator, validated_params.permissions, validated_params.flow_rate_allowance, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.updateFlowOperatorPermissions(
            validated_params.super_token, validated_params.flow_operator, validated_params.permissions, validated_params.flow_rate_allowance)
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def authorize_flow_operator_with_full_control(self, super_token: HexAddress, flow_operator: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Give flow operator full control - max flow rate and create/update/delete permissions.

            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = FullControlParams(
            super_token, flow_operator, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='authorizeFlowOperatorWithFullControl', args=[
            validated_params.super_token, validated_params.flow_operator, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.grantPermissions(
            validated_params.super_token, validated_params.flow_operator)
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def revoke_flow_operator_with_full_control(self, super_token: HexAddress, flow_operator: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Revoke flow operator control - set flow rate to 0 with no permissions.

            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = FullControlParams(
            super_token, flow_operator, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='revokeFlowOperatorWithFullControl', args=[
            validated_params.super_token, validated_params.flow_operator, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        forwarder_call: ContractFunction = self.forwarder.functions.revokePermissions(
            validated_params.super_token, validated_params.flow_operator)
        return self._get_call_agreement_operation(call_agreement_operation, forwarder_call, validated_params.should_use_call_agreement)

    def create_flow_by_operator(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, flow_rate: int, user_data: Optional[HexStr] = None,  should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Create a flow as an operator

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param flow_rate - flow rate for the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = CreateFlowByOperatorParams(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='createFlowByOperator', args=[
            validated_params.super_token, validated_params.sender, validated_params.receiver, validated_params.flow_rate, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, validated_params.user_data or "0x")
        create_flow_operation = self.create_flow(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        return self._get_call_agreement_operation(call_agreement_operation, create_flow_operation.forwarder_call, validated_params.should_use_call_agreement)

    def update_flow_by_operator(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, flow_rate: int, user_data: Optional[HexStr] = None,  should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Update a flow as an operator

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param flow_rate - flow rate for the flow
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = UpdateFlowByOperatorParams(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='updateFlowByOperator', args=[
            validated_params.super_token, validated_params.sender, validated_params.receiver, validated_params.flow_rate, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, "0x")
        update_flow_operation = self.update_flow(
            sender, receiver, super_token, flow_rate, user_data, should_use_call_agreement)
        return self._get_call_agreement_operation(call_agreement_operation, update_flow_operation.forwarder_call, validated_params.should_use_call_agreement)

    def delete_flow_by_operator(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        """
            Delete a flow as an operator

            @param sender - sender of the flow
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param user_data(Optional) - Extra data provided
            @param should_use_call_agreement(Optional) - whether or not to use the host contract
            @returns - Operation
        """
        validated_params = DeleteFlowParams(
            sender, receiver, super_token, user_data, should_use_call_agreement)
        calldata = self.contract.encodeABI(fn_name='deleteFlowByOperator', args=[
            validated_params.super_token, validated_params.sender, validated_params.receiver, "0x"])
        call_agreement_operation = self.host.call_agreement(
            self.contract.address, calldata, "0x")
        delete_flow_operation = self.delete_flow(
            sender, receiver, super_token, user_data, should_use_call_agreement)
        return self._get_call_agreement_operation(call_agreement_operation, delete_flow_operation.forwarder_call, validated_params.should_use_call_agreement)

    def _get_call_agreement_operation(self, call_agreement_operation: Operation, forwarder_call: Optional[ContractFunction] = None, should_use_call_agreement: Optional[bool] = None) -> Operation:
        if should_use_call_agreement == True:
            return call_agreement_operation
        else:
            return Operation(self.rpc, call_agreement_operation.agreement_call, call_agreement_operation.type, forwarder_call)



from typing import Optional, Type

from web3 import Web3
from web3.contract.contract import ContractFunction
from web3.middleware import geth_poa_middleware

from .constants import HOST_ABI
from .operation import Operation
from .types import BatchOperationType


class Host:

    def __init__(self, rpc: str, host_address: str) -> None:
        self.rpc = rpc
        web3 = Web3(Web3.HTTPProvider(rpc))
        web3.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.contract = web3.eth.contract(
            address=host_address, abi=HOST_ABI)

    def call_agreement(self, agreement_address: str, calldata: str, user_data: str) -> Operation:
        try:
            agreement_call: ContractFunction = self.contract.functions.callAgreement(
                agreement_address, calldata, user_data)
            return Operation(self.rpc, agreement_call, BatchOperationType.SUPERFLUID_CALL_AGREEMENT)
        except Exception as e:
            raise e

    def call_app_action() -> Operation:
        pass



networks = '''[
    {
        "name": "eth-goerli",
        "isTestnet": true,
        "networkId": 5,
        "chainId": 5,
        "shortName": "goerli",
        "uppercaseName": "ETH_GOERLI",
        "humanReadableName": "Goerli",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x3710AB3fDE2B61736B8BB0CE845D6c61F667a78E",
            "host": "0x22ff293e14F1EC3A09B137e9e06084AFd63adDF9",
            "governance": "0x3a648764a6d66440ca096343937c711a7ac1b1e9",
            "cfaV1": "0xEd6BcbF6907D4feEEe8a8875543249bEa9D308E8",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0xfDdcdac21D64B639546f3Ce2868C7EF06036990c",
            "superTokenFactory": "0x94f26B4c8AD12B18c12f38E878618f7664bdcCE2",
            "superfluidLoader": "0x74d860243Ff08A243d5485899f343117EbDa6eA8",
            "toga": "0xa54FC15FC75693447d70a57262F37a70B614721b",
            "flowScheduler": "0xf428308b426D7cD7Ad8eBE549d750f31C8E060Ca",
            "vestingScheduler": "0xF9240F930d847F70ad900aBEE8949F25649Bf24a"
        },
        "startBlockV1": 3550000,
        "logsQueryRange": 10000,
        "explorer": "https://goerli.etherscan.io",
        "subgraphV1": {
            "name": "protocol-v1-goerli",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-goerli"
        },
        "publicRPCs": [
            "https://rpc.ankr.com/eth_goerli",
            "https://goerli.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161"
        ]
    },
    {
        "name": "polygon-mumbai",
        "isTestnet": true,
        "networkId": 80001,
        "chainId": 80001,
        "shortName": "mumbai",
        "uppercaseName": "POLYGON_MUMBAI",
        "humanReadableName": "Polygon Mumbai",
        "nativeTokenSymbol": "MATIC",
        "contractsV1": {
            "resolver": "0x8C54C83FbDe3C59e59dd6E324531FB93d4F504d3",
            "host": "0xEB796bdb90fFA0f28255275e16936D25d3418603",
            "governance": "0x2637eA93EE5cd887ff9AC98185eA67Bd70C5f62e",
            "cfaV1": "0x49e565Ed1bdc17F3d220f72DF0857C26FA83F873",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x804348D4960a61f2d5F9ce9103027A3E849E09b8",
            "superTokenFactory": "0xB798553db6EB3D3C56912378409370145E97324B",
            "superfluidLoader": "0x0d56ED56b63382B0FC964490feB9AE438B6B4b79",
            "toga": "0x38DD80876DBA048d0050D28828522c313967D073",
            "superSpreader": "0x74CDF863b00789c29734F8dFd9F83423Bc55E4cE",
            "flowScheduler": "0x59A3Ba9d34c387FB70b4f4e4Fbc9eD7519194139",
            "vestingScheduler": "0x3962EE56c9f7176215D149938BA685F91aBB633B"
        },
        "startBlockV1": 8100000,
        "logsQueryRange": 10000,
        "explorer": "https://mumbai.polygonscan.com",
        "subgraphV1": {
            "name": "protocol-v1-mumbai",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-mumbai"
        },
        "publicRPCs": [
            "https://rpc.ankr.com/polygon_mumbai",
            "https://matic-mumbai.chainstacklabs.com"
        ]
    },
    {
        "name": "optimism-goerli",
        "isTestnet": true,
        "networkId": 420,
        "chainId": 420,
        "shortName": "opgoerli",
        "uppercaseName": "OPTIMISM_GOERLI",
        "humanReadableName": "Optimism Goerli",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136",
            "host": "0xE40983C2476032A0915600b9472B3141aA5B5Ba9",
            "governance": "0x777Be25F9fdcA87e8a0E06Ad4be93d65429FCb9f",
            "cfaV1": "0xff48668fa670A85e55A7a822b352d5ccF3E7b18C",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x96215257F2FcbB00135578f766c0449d239bd92F",
            "superTokenFactory": "0xfafe31cf998Df4e5D8310B03EBa8fb5bF327Eaf5",
            "superfluidLoader": "0x5Bb5908dcCC9Bb0fC39a78CfDf9e47B4C08E9521"
        },
        "startBlockV1": 340000,
        "logsQueryRange": 50000,
        "explorer": "https://goerli-optimism.etherscan.io",
        "subgraphV1": {
            "name": "protocol-v1-optimism-goerli",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-optimism-goerli"
        }
    },
    {
        "name": "arbitrum-goerli",
        "isTestnet": true,
        "networkId": 421613,
        "chainId": 421613,
        "shortName": "arbgoerli",
        "uppercaseName": "ARBITRUM_GOERLI",
        "humanReadableName": "Arbitrum Goerli",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136",
            "host": "0xE40983C2476032A0915600b9472B3141aA5B5Ba9",
            "governance": "0x777Be25F9fdcA87e8a0E06Ad4be93d65429FCb9f",
            "cfaV1": "0xff48668fa670A85e55A7a822b352d5ccF3E7b18C",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x96215257F2FcbB00135578f766c0449d239bd92F",
            "superTokenFactory": "0xfafe31cf998Df4e5D8310B03EBa8fb5bF327Eaf5",
            "superfluidLoader": "0x5Bb5908dcCC9Bb0fC39a78CfDf9e47B4C08E9521"
        },
        "startBlockV1": 93000,
        "logsQueryRange": 50000,
        "explorer": "https://goerli.arbiscan.io",
        "subgraphV1": {
            "name": "protocol-v1-arbitrum-goerli",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-arbitrum-goerli"
        }
    },
    {
        "name": "avalanche-fuji",
        "isTestnet": true,
        "networkId": 43113,
        "chainId": 43113,
        "shortName": "fuji",
        "uppercaseName": "AVALANCHE_FUJI",
        "humanReadableName": "Avalanche Fuji",
        "nativeTokenSymbol": "AVAX",
        "contractsV1": {
            "resolver": "0xf0ec6A8842Ca72Aec8A4D4573E731242389e18A8",
            "host": "0x85Fe79b998509B77BF10A8BD4001D58475D29386",
            "governance": "0xA55632254Bc9F739bDe7191c8a4510aDdae3ef6D",
            "cfaV1": "0x16843ac25Ccc58Aa7960ba05f61cBB17b36b130A",
            "cfaV1Forwarder": "0x2CDd45c5182602a36d391F7F16DD9f8386C3bD8D",
            "idaV1": "0xA44dEC7A0Dde1a56AeDe4143C1ef89cf5d956782",
            "superTokenFactory": "0x1C92042426B6bAAe497bEf461B6d8342D03aEc92",
            "superfluidLoader": "0x96C3C2d23d143301cF363a02cB7fe3596d2834d7"
        },
        "startBlockV1": 3220000,
        "logsQueryRange": 50000,
        "explorer": "https://testnet.snowtrace.io",
        "subgraphV1": {
            "name": "protocol-v1-avalanche-fuji",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-avalanche-fuji"
        }
    },
    {
        "name": "eth-sepolia",
        "isTestnet": true,
        "networkId": 11155111,
        "chainId": 11155111,
        "shortName": "sepolia",
        "uppercaseName": "ETH_SEPOLIA",
        "humanReadableName": "Sepolia",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x6813edE4E78ecb830d380d0F7F684c12aAc95F02",
            "host": "0x109412E3C84f0539b43d39dB691B08c90f58dC7c",
            "governance": "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136",
            "cfaV1": "0x6836F23d6171D74Ef62FcF776655aBcD2bcd62Ef",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x9358C7dCCc6B8CA6F526311e8ac266F8C861B7ea",
            "superTokenFactory": "0x254C2e152E8602839D288A7bccdf3d0974597193",
            "superfluidLoader": "0x554c06487bEc8c890A0345eb05a5292C1b1017Bd"
        },
        "startBlockV1": 3322400,
        "logsQueryRange": 10000,
        "explorer": "https://sepolia.etherscan.io",
        "subgraphV1": {
            "name": "protocol-v1-eth-sepolia",
            "satsumaEndpoint": "https://subgraph.satsuma-prod.com/c5br3jaVlJI6/superfluid/eth-sepolia/api"
        },
        "publicRPCs": [
            "https://rpc.sepolia.org",
            "https://ethereum-sepolia.blockpi.network/v1/rpc/public"
        ]
    },
    {
        "name": "xdai-mainnet",
        "isTestnet": false,
        "networkId": 100,
        "chainId": 100,
        "shortName": "xdai",
        "uppercaseName": "XDAI_MAINNET",
        "humanReadableName": "Gnosis Chain",
        "nativeTokenSymbol": "xDAI",
        "contractsV1": {
            "resolver": "0xD2009765189164b495c110D61e4D301729079911",
            "host": "0x2dFe937cD98Ab92e59cF3139138f18c823a4efE7",
            "governance": "0xaCc7380323681fdb8a0B9F2FE7d69dDFf0664478",
            "cfaV1": "0xEbdA4ceF883A7B12c4E669Ebc58927FBa8447C7D",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x7888ac96F987Eb10E291F34851ae0266eF912081",
            "superTokenFactory": "0x23410e2659380784498509698ed70E414D384880",
            "superfluidLoader": "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136",
            "toga": "0xb7DE52F4281a7a276E18C40F94cd93159C4A2d22",
            "batchLiquidator": "0x27636F8E129cdd4ccA0F30E2b4C116DDaC773bE5",
            "superSpreader": "0x74CDF863b00789c29734F8dFd9F83423Bc55E4cE",
            "flowScheduler": "0x9cC7fc484fF588926149577e9330fA5b2cA74336",
            "vestingScheduler": "0x0170FFCC75d178d426EBad5b1a31451d00Ddbd0D",
            "wrapManager": "0x7a2899D179a8F205C8EDAd2e52954cA5f6d48D1A",
            "wrapStrategy": "0xc3B7f0b221a002fE8Fc93b4Ef9BB6362950510F2"
        },
        "startBlockV1": 14820000,
        "logsQueryRange": 20000,
        "explorer": "https://gnosisscan.io",
        "subgraphV1": {
            "name": "protocol-v1-xdai",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-xdai",
            "satsumaEndpoint": "https://subgraph.satsuma-prod.com/c5br3jaVlJI6/superfluid/xdai/api"
        },
        "publicRPCs": [
            "https://rpc.gnosischain.com",
            "https://gnosischain-rpc.gateway.pokt.network"
        ],
        "coinGeckoId": "xdai"
    },
    {
        "name": "polygon-mainnet",
        "isTestnet": false,
        "networkId": 137,
        "chainId": 137,
        "shortName": "matic",
        "uppercaseName": "POLYGON_MAINNET",
        "humanReadableName": "Polygon",
        "nativeTokenSymbol": "MATIC",
        "contractsV1": {
            "resolver": "0xE0cc76334405EE8b39213E620587d815967af39C",
            "host": "0x3E14dC1b13c488a8d5D310918780c983bD5982E7",
            "governance": "0x3AD3f7A0965Ce6f9358AD5CCE86Bc2b05F1EE087",
            "cfaV1": "0x6EeE6060f715257b970700bc2656De21dEdF074C",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0xB0aABBA4B2783A72C52956CDEF62d438ecA2d7a1",
            "superTokenFactory": "0x2C90719f25B10Fc5646c82DA3240C76Fa5BcCF34",
            "superfluidLoader": "0x15F0Ca26781C3852f8166eD2ebce5D18265cceb7",
            "toga": "0x6AEAeE5Fd4D05A741723D752D30EE4D72690A8f7",
            "batchLiquidator": "0xA6Cdb472e7E22Bf30ae6fB752E4a13eBF3c12165",
            "flowScheduler": "0x55F7758dd99d5e185f4CC08d4Ad95B71f598264D",
            "vestingScheduler": "0xcFE6382B33F2AdaFbE46e6A26A88E0182ae32b0c"
        },
        "startBlockV1": 11650500,
        "logsQueryRange": 10000,
        "explorer": "https://polygonscan.com",
        "subgraphV1": {
            "name": "protocol-v1-matic",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-matic",
            "satsumaEndpoint": "https://subgraph.satsuma-prod.com/c5br3jaVlJI6/superfluid/matic/api"
        },
        "publicRPCs": [
            "https://polygon-rpc.com",
            "https://poly-rpc.gateway.pokt.network"
        ],
        "coinGeckoId": "polygon-pos"
    },
    {
        "name": "optimism-mainnet",
        "isTestnet": false,
        "networkId": 10,
        "chainId": 10,
        "shortName": "optimism",
        "uppercaseName": "OPTIMISM_MAINNET",
        "humanReadableName": "Optimism",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x743B5f46BC86caF41bE4956d9275721E0531B186",
            "host": "0x567c4B141ED61923967cA25Ef4906C8781069a10",
            "governance": "0x0170FFCC75d178d426EBad5b1a31451d00Ddbd0D",
            "cfaV1": "0x204C6f131bb7F258b2Ea1593f5309911d8E458eD",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0xc4ce5118C3B20950ee288f086cb7FC166d222D4c",
            "superTokenFactory": "0x8276469A443D5C6B7146BED45e2abCaD3B6adad9",
            "superfluidLoader": "0x8E310ce29Ab7Fa2878944A65BB0eaF97B1853d40",
            "toga": "0xA3c8502187fD7a7118eAD59dc811281448946C8f",
            "batchLiquidator": "0x36Df169DBf5CE3c6f58D46f0addeF58F01381232",
            "flowScheduler": "0x55c8fc400833eEa791087cF343Ff2409A39DeBcC",
            "vestingScheduler": "0x65377d4dfE9c01639A41952B5083D58964782892"
        },
        "startBlockV1": 4300000,
        "logsQueryRange": 50000,
        "explorer": "https://optimistic.etherscan.io",
        "subgraphV1": {
            "name": "protocol-v1-optimism-mainnet",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-optimism-mainnet"
        },
        "publicRPCs": [
            "https://mainnet.optimism.io",
            "https://rpc.ankr.com/optimism"
        ],
        "coinGeckoId": "optimistic-ethereum"
    },
    {
        "name": "arbitrum-one",
        "isTestnet": false,
        "networkId": 42161,
        "chainId": 42161,
        "shortName": "arbone",
        "uppercaseName": "ARBITRUM_ONE",
        "humanReadableName": "Arbitrum One",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x609b9d9d6Ee9C3200745A79B9d3398DBd63d509F",
            "host": "0xCf8Acb4eF033efF16E8080aed4c7D5B9285D2192",
            "governance": "0x0170FFCC75d178d426EBad5b1a31451d00Ddbd0D",
            "cfaV1": "0x731FdBB12944973B500518aea61942381d7e240D",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x2319C7e07EB063340D2a0E36709B0D65fda75986",
            "superTokenFactory": "0x1C21Ead77fd45C84a4c916Db7A6635D0C6FF09D6",
            "superfluidLoader": "0xB99cA401e10D277345BcFb281AC148a2e16Db466",
            "toga": "0xFC63B7C762B10670Eda15cF3ca3970bCDB28C9eF",
            "batchLiquidator": "0x6C66e5c5D201A753ff497F2e9eC5D545631854d0",
            "flowScheduler": "0x3fA8B653F9abf91428800C0ba0F8D145a71F97A1",
            "vestingScheduler": "0x55c8fc400833eEa791087cF343Ff2409A39DeBcC"
        },
        "startBlockV1": 7600000,
        "logsQueryRange": 50000,
        "explorer": "https://arbiscan.io",
        "subgraphV1": {
            "name": "protocol-v1-arbitrum-one",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-arbitrum-one"
        },
        "publicRPCs": [
            "https://arb1.arbitrum.io/rpc",
            "https://arbitrum.blockpi.network/v1/rpc/public"
        ],
        "coinGeckoId": "arbitrum-one"
    },
    {
        "name": "avalanche-c",
        "isTestnet": false,
        "networkId": 43114,
        "chainId": 43114,
        "shortName": "avalanche",
        "uppercaseName": "AVALANCHE_C",
        "humanReadableName": "Avalanche C",
        "nativeTokenSymbol": "AVAX",
        "contractsV1": {
            "resolver": "0x24a3F04F70B7f07B9673EadD3e146391BcfEa5c1",
            "host": "0x60377C7016E4cdB03C87EF474896C11cB560752C",
            "governance": "0xF74390BabA510ec2fE196c2e02B037380d7a6F12",
            "cfaV1": "0x6946c5B38Ffea373b0a2340b4AEf0De8F6782e58",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x1fA9fFe8Db73F701454B195151Db4Abb18423cf2",
            "superTokenFactory": "0x464AADdBB2B80f3Cb666522EB7381bE610F638b4",
            "superfluidLoader": "0x2601E69a2D16C8Ccc8aEe8cE9F59d24a59986B9E",
            "toga": "0x3D9A67D5ec1E72CEcA8157e028855056786b6159",
            "batchLiquidator": "0xdddaD64A9Fe7709A729C4a5428617e369278e0b6",
            "flowScheduler": "0xF7AfF590E9DE493D7ACb421Fca7f1E35C1ad4Ce5",
            "vestingScheduler": "0x3fA8B653F9abf91428800C0ba0F8D145a71F97A1"
        },
        "startBlockV1": 14700000,
        "logsQueryRange": 50000,
        "explorer": "https://snowtrace.io",
        "subgraphV1": {
            "name": "protocol-v1-avalanche-c",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-avalanche-c"
        },
        "publicRPCs": [
            "https://api.avax.network/ext/bc/C/rpc",
            "https://rpc.ankr.com/avalanche"
        ],
        "coinGeckoId": "avalanche"
    },
    {
        "name": "bsc-mainnet",
        "isTestnet": false,
        "networkId": 56,
        "chainId": 56,
        "shortName": "bsc",
        "uppercaseName": "BSC_MAINNET",
        "humanReadableName": "BNB Smart Chain",
        "nativeTokenSymbol": "BNB",
        "contractsV1": {
            "resolver": "0x69604aA4e9e8BF44A73C680997205Edb03A92E41",
            "host": "0xd1e2cFb6441680002Eb7A44223160aB9B67d7E6E",
            "governance": "0xee07D9fce4Cf2a891BC979E9d365929506C2982f",
            "cfaV1": "0x49c38108870e74Cb9420C0991a85D3edd6363F75",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x594ed9Cd773584B645aC1F5B11020d3b32cDF07d",
            "superTokenFactory": "0x8bde47397301F0Cd31b9000032fD517a39c946Eb",
            "superfluidLoader": "0x3C26e5bEbc68EaAf4efcd88F8E5A6Aeb77b65579",
            "toga": "0xFCD84210f5d51Cd40a30443d44d6A5500d5D10dF",
            "batchLiquidator": "0x5487d078CA8933e83d91d5E7AFBe3A7bfC3412d6",
            "flowScheduler": "0x2f9e2A2A59405682d4F86779275CF5525AD7eC2B",
            "vestingScheduler": "0x9B91c27f78376383003C6A12Ad12B341d016C5b9"
        },
        "startBlockV1": 18800000,
        "logsQueryRange": 5000,
        "explorer": "https://bscscan.com",
        "subgraphV1": {
            "name": "protocol-v1-bsc-mainnet",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-bsc-mainnet"
        },
        "publicRPCs": [
            "https://bscrpc.com",
            "https://bsc-dataseed.binance.org"
        ],
        "coinGeckoId": "binance-smart-chain"
    },
    {
        "name": "eth-mainnet",
        "isTestnet": false,
        "networkId": 1,
        "chainId": 1,
        "shortName": "mainnet",
        "uppercaseName": "ETH_MAINNET",
        "humanReadableName": "Ethereum",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0xeE4cD028f5fdaAdeA99f8fc38e8bA8A57c90Be53",
            "host": "0x4E583d9390082B65Bef884b629DFA426114CED6d",
            "governance": "0xe2E14e2C4518cB06c32Cd0818B4C01f53E1Ba653",
            "cfaV1": "0x2844c1BBdA121E9E43105630b9C8310e5c72744b",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0xbCF9cfA8Da20B591790dF27DE65C1254Bf91563d",
            "superTokenFactory": "0x0422689cc4087b6B7280e0a7e7F655200ec86Ae1",
            "superfluidLoader": "0x9775EEBdfF8AE845e7Ad3d1D04B85e6c6d284aCB",
            "toga": "0x8B5a2CF69a56d7F8Fa027edcA23594cdDF544dDc",
            "batchLiquidator": "0x554c06487bEc8c890A0345eb05a5292C1b1017Bd",
            "flowScheduler": "0xAA0cD305eD020137E302CeCede7b18c0A05aCCDA",
            "vestingScheduler": "0x39D5cBBa9adEBc25085a3918d36D5325546C001B"
        },
        "startBlockV1": 15870000,
        "logsQueryRange": 10000,
        "explorer": "https://etherscan.io",
        "subgraphV1": {
            "name": "protocol-v1-eth-mainnet",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-eth-mainnet",
            "satsumaEndpoint": "https://subgraph.satsuma-prod.com/c5br3jaVlJI6/superfluid/eth-mainnet/api"
        },
        "publicRPCs": [
            "https://cloudflare-eth.com",
            "https://eth-rpc.gateway.pokt.network"
        ],
        "coinGeckoId": "ethereum"
    },
    {
        "name": "celo-mainnet",
        "isTestnet": false,
        "networkId": 42220,
        "chainId": 42220,
        "shortName": "celo",
        "uppercaseName": "CELO_MAINNET",
        "humanReadableName": "Celo",
        "nativeTokenSymbol": "CELO",
        "contractsV1": {
            "resolver": "0x05eE721BD4D803d6d477Aa7607395452B65373FF",
            "host": "0xA4Ff07cF81C02CFD356184879D953970cA957585",
            "governance": "0x0170FFCC75d178d426EBad5b1a31451d00Ddbd0D",
            "cfaV1": "0x9d369e78e1a682cE0F8d9aD849BeA4FE1c3bD3Ad",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x26747Fe93fAC8bF28E1e24A558a2bC7E4d9846cA",
            "superTokenFactory": "0x36be86dEe6BC726Ed0Cbd170ccD2F21760BC73D9",
            "superfluidLoader": "0xAd6e6849d8eEa62AF9271808afb726610fB451a6",
            "toga": "0x9bCa3a623e7b2e248510d88B2894F54898d88F91",
            "batchLiquidator": "0x21d4E9fbB9DB742E6ef4f29d189a7C18B0b59136"
        },
        "startBlockV1": 16393000,
        "logsQueryRange": 20000,
        "explorer": "https://celoscan.io",
        "subgraphV1": {
            "name": "protocol-v1-celo-mainnet",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-celo-mainnet"
        },
        "publicRPCs": [
            "https://forno.celo.org",
            "https://rpc.ankr.com/celo"
        ],
        "coinGeckoId": "celo"
    },
    {
        "name": "base-mainnet",
        "isTestnet": false,
        "networkId": 8453,
        "chainId": 8453,
        "shortName": "base",
        "uppercaseName": "base-mainnet",
        "humanReadableName": "Base Mainnet",
        "nativeTokenSymbol": "ETH",
        "contractsV1": {
            "resolver": "0x6a214c324553F96F04eFBDd66908685525Da0E0d",
            "host": "0x4C073B3baB6d8826b8C5b229f3cfdC1eC6E47E74",
            "governance": "0x3a648764a6d66440ca096343937c711a7ac1b1e9",
            "cfaV1": "0x19ba78B9cDB05A877718841c574325fdB53601bb",
            "cfaV1Forwarder": "0xcfA132E353cB4E398080B9700609bb008eceB125",
            "idaV1": "0x66DF3f8e14CF870361378d8F61356D15d9F425C4",
            "superTokenFactory": "0xe20B9a38E0c96F61d1bA6b42a61512D56Fea1Eb3",
            "superfluidLoader": "0x9F9EEd9Bbc38E9e0514fD9dFcc0Bca9869A9c534",
            "toga": "0xA87F76e99f6C8Ff8996d14f550ceF47f193D9A09",
            "flowScheduler": "0xC72CEd15204d02183c83fEbb918b183E400811Ee",
            "vestingScheduler": "0xDF92D0E6Bcb9385FDe99aD21Ff5e47Fb47E3c6b2"
        },
        "startBlockV1": 18081759,
        "logsQueryRange": 10000,
        "explorer": "https://base.blockscout.com/",
        "subgraphV1": {
            "name": "protocol-v1-celo-mainnet",
            "hostedEndpoint": "https://api.thegraph.com/subgraphs/name/superfluid-finance/protocol-v1-celo-mainnet"
        },
        "publicRPCs": [
            "https://mainnet.base.org"
        ]
    }
]'''



cfa_v1_forwarder = '''[
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluid",
        "name": "host",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  { "inputs": [], "name": "CFA_FWD_INVALID_FLOW_RATE", "type": "error" },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" },
      { "internalType": "bytes", "name": "userData", "type": "bytes" }
    ],
    "name": "createFlow",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "bytes", "name": "userData", "type": "bytes" }
    ],
    "name": "deleteFlow",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "getAccountFlowInfo",
    "outputs": [
      { "internalType": "uint256", "name": "lastUpdated", "type": "uint256" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "getAccountFlowrate",
    "outputs": [
      { "internalType": "int96", "name": "flowrate", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "int96", "name": "flowrate", "type": "int96" }
    ],
    "name": "getBufferAmountByFlowrate",
    "outputs": [
      { "internalType": "uint256", "name": "bufferAmount", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" }
    ],
    "name": "getFlowInfo",
    "outputs": [
      { "internalType": "uint256", "name": "lastUpdated", "type": "uint256" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "flowOperator", "type": "address" }
    ],
    "name": "getFlowOperatorPermissions",
    "outputs": [
      { "internalType": "uint8", "name": "permissions", "type": "uint8" },
      { "internalType": "int96", "name": "flowrateAllowance", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" }
    ],
    "name": "getFlowrate",
    "outputs": [
      { "internalType": "int96", "name": "flowrate", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" }
    ],
    "name": "grantPermissions",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" }
    ],
    "name": "revokePermissions",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" }
    ],
    "name": "setFlowrate",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" }
    ],
    "name": "setFlowrateFrom",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowrate", "type": "int96" },
      { "internalType": "bytes", "name": "userData", "type": "bytes" }
    ],
    "name": "updateFlow",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "uint8", "name": "permissions", "type": "uint8" },
      { "internalType": "int96", "name": "flowrateAllowance", "type": "int96" }
    ],
    "name": "updateFlowOperatorPermissions",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
'''

cfa_v1 = '''[
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluid",
        "name": "host",
        "type": "address"
      },
      {
        "internalType": "contract IConstantFlowAgreementHook",
        "name": "_hookAddress",
        "type": "address"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "ALREADY_EXISTS",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "APP_RULE",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "CFA_ACL_FLOW_RATE_ALLOWANCE_EXCEEDED",
    "type": "error"
  },
  { "inputs": [], "name": "CFA_ACL_NO_NEGATIVE_ALLOWANCE", "type": "error" },
  { "inputs": [], "name": "CFA_ACL_NO_SENDER_CREATE", "type": "error" },
  { "inputs": [], "name": "CFA_ACL_NO_SENDER_FLOW_OPERATOR", "type": "error" },
  { "inputs": [], "name": "CFA_ACL_NO_SENDER_UPDATE", "type": "error" },
  {
    "inputs": [],
    "name": "CFA_ACL_OPERATOR_NO_CREATE_PERMISSIONS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "CFA_ACL_OPERATOR_NO_DELETE_PERMISSIONS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "CFA_ACL_OPERATOR_NO_UPDATE_PERMISSIONS",
    "type": "error"
  },
  { "inputs": [], "name": "CFA_ACL_UNCLEAN_PERMISSIONS", "type": "error" },
  { "inputs": [], "name": "CFA_DEPOSIT_TOO_BIG", "type": "error" },
  { "inputs": [], "name": "CFA_FLOW_RATE_TOO_BIG", "type": "error" },
  { "inputs": [], "name": "CFA_INVALID_FLOW_RATE", "type": "error" },
  { "inputs": [], "name": "CFA_NON_CRITICAL_SENDER", "type": "error" },
  { "inputs": [], "name": "CFA_NO_SELF_FLOW", "type": "error" },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "DOES_NOT_EXIST",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "INSUFFICIENT_BALANCE",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "ONLY_HOST",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "OUT_OF_GAS",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "ZERO_ADDRESS",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bytes32",
        "name": "uuid",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "codeAddress",
        "type": "address"
      }
    ],
    "name": "CodeUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "flowOperator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "permissions",
        "type": "uint8"
      },
      {
        "indexed": false,
        "internalType": "int96",
        "name": "flowRateAllowance",
        "type": "int96"
      }
    ],
    "name": "FlowOperatorUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "sender",
        "type": "address"
      },
      {
        "indexed": true,
        "internalType": "address",
        "name": "receiver",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "int96",
        "name": "flowRate",
        "type": "int96"
      },
      {
        "indexed": false,
        "internalType": "int256",
        "name": "totalSenderFlowRate",
        "type": "int256"
      },
      {
        "indexed": false,
        "internalType": "int256",
        "name": "totalReceiverFlowRate",
        "type": "int256"
      },
      {
        "indexed": false,
        "internalType": "bytes",
        "name": "userData",
        "type": "bytes"
      }
    ],
    "name": "FlowUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "address",
        "name": "flowOperator",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "deposit",
        "type": "uint256"
      }
    ],
    "name": "FlowUpdatedExtension",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "CFA_HOOK_GAS_LIMIT",
    "outputs": [{ "internalType": "uint64", "name": "", "type": "uint64" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "DEFAULT_MINIMUM_DEPOSIT",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "MAXIMUM_DEPOSIT",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "MAXIMUM_FLOW_RATE",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "agreementType",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "authorizeFlowOperatorWithFullControl",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "castrate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "constantFlowAgreementHook",
    "outputs": [
      {
        "internalType": "contract IConstantFlowAgreementHook",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "createFlow",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "createFlowByOperator",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "deleteFlow",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "deleteFlowByOperator",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "getAccountFlowInfo",
    "outputs": [
      { "internalType": "uint256", "name": "timestamp", "type": "uint256" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getCodeAddress",
    "outputs": [
      { "internalType": "address", "name": "codeAddress", "type": "address" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "int96", "name": "flowRate", "type": "int96" }
    ],
    "name": "getDepositRequiredForFlowRate",
    "outputs": [
      { "internalType": "uint256", "name": "deposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" }
    ],
    "name": "getFlow",
    "outputs": [
      { "internalType": "uint256", "name": "timestamp", "type": "uint256" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "bytes32", "name": "flowId", "type": "bytes32" }
    ],
    "name": "getFlowByID",
    "outputs": [
      { "internalType": "uint256", "name": "timestamp", "type": "uint256" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "flowOperator", "type": "address" }
    ],
    "name": "getFlowOperatorData",
    "outputs": [
      {
        "internalType": "bytes32",
        "name": "flowOperatorId",
        "type": "bytes32"
      },
      { "internalType": "uint8", "name": "permissions", "type": "uint8" },
      { "internalType": "int96", "name": "flowRateAllowance", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "bytes32", "name": "flowOperatorId", "type": "bytes32" }
    ],
    "name": "getFlowOperatorDataByID",
    "outputs": [
      { "internalType": "uint8", "name": "permissions", "type": "uint8" },
      { "internalType": "int96", "name": "flowRateAllowance", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" }
    ],
    "name": "getMaximumFlowRateFromDeposit",
    "outputs": [
      { "internalType": "int96", "name": "flowRate", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "getNetFlow",
    "outputs": [
      { "internalType": "int96", "name": "flowRate", "type": "int96" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" },
      { "internalType": "uint256", "name": "timestamp", "type": "uint256" }
    ],
    "name": "isPatricianPeriod",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" }
    ],
    "name": "isPatricianPeriodNow",
    "outputs": [
      {
        "internalType": "bool",
        "name": "isCurrentlyPatricianPeriod",
        "type": "bool"
      },
      { "internalType": "uint256", "name": "timestamp", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "proxiableUUID",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "account", "type": "address" },
      { "internalType": "uint256", "name": "time", "type": "uint256" }
    ],
    "name": "realtimeBalanceOf",
    "outputs": [
      { "internalType": "int256", "name": "dynamicBalance", "type": "int256" },
      { "internalType": "uint256", "name": "deposit", "type": "uint256" },
      { "internalType": "uint256", "name": "owedDeposit", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "revokeFlowOperatorWithFullControl",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "newAddress", "type": "address" }
    ],
    "name": "updateCode",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "updateFlow",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "sender", "type": "address" },
      { "internalType": "address", "name": "receiver", "type": "address" },
      { "internalType": "int96", "name": "flowRate", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "updateFlowByOperator",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "uint8", "name": "permissions", "type": "uint8" },
      { "internalType": "int96", "name": "flowRateAllowance", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "updateFlowOperatorPermissions",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "int96", "name": "addedFlowRateAllowance", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "increaseFlowRateAllowance",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidToken",
        "name": "token",
        "type": "address"
      },
      { "internalType": "address", "name": "flowOperator", "type": "address" },
      { "internalType": "int96", "name": "subtractedFlowRateAllowance", "type": "int96" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "decreaseFlowRateAllowance",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  }
]
'''

host = '''[
  {
    "inputs": [
      { "internalType": "bool", "name": "nonUpgradable", "type": "bool" },
      {
        "internalType": "bool",
        "name": "appWhiteListingEnabled",
        "type": "bool"
      }
    ],
    "stateMutability": "nonpayable",
    "type": "constructor"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "ALREADY_EXISTS",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "APP_RULE",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "DOES_NOT_EXIST",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_AGREEMENT_CALLBACK_IS_NOT_ACTION",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_CALL_AGREEMENT_WITH_CTX_FROM_WRONG_ADDRESS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_CALL_APP_ACTION_WITH_CTX_FROM_WRONG_ADDRESS",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_CANNOT_DOWNGRADE_TO_NON_UPGRADEABLE",
    "type": "error"
  },
  { "inputs": [], "name": "HOST_INVALID_CONFIG_WORD", "type": "error" },
  {
    "inputs": [],
    "name": "HOST_INVALID_OR_EXPIRED_SUPER_APP_REGISTRATION_KEY",
    "type": "error"
  },
  { "inputs": [], "name": "HOST_MAX_256_AGREEMENTS", "type": "error" },
  { "inputs": [], "name": "HOST_NON_UPGRADEABLE", "type": "error" },
  {
    "inputs": [],
    "name": "HOST_NON_ZERO_LENGTH_PLACEHOLDER_CTX",
    "type": "error"
  },
  { "inputs": [], "name": "HOST_NOT_A_SUPER_APP", "type": "error" },
  {
    "inputs": [],
    "name": "HOST_NO_APP_REGISTRATION_PERMISSIONS",
    "type": "error"
  },
  { "inputs": [], "name": "HOST_ONLY_GOVERNANCE", "type": "error" },
  { "inputs": [], "name": "HOST_RECEIVER_IS_NOT_SUPER_APP", "type": "error" },
  { "inputs": [], "name": "HOST_SENDER_IS_NOT_SUPER_APP", "type": "error" },
  {
    "inputs": [],
    "name": "HOST_SOURCE_APP_NEEDS_HIGHER_APP_LEVEL",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_SUPER_APP_ALREADY_REGISTERED",
    "type": "error"
  },
  { "inputs": [], "name": "HOST_SUPER_APP_IS_JAILED", "type": "error" },
  {
    "inputs": [],
    "name": "HOST_UNAUTHORIZED_SUPER_APP_FACTORY",
    "type": "error"
  },
  {
    "inputs": [],
    "name": "HOST_UNKNOWN_BATCH_CALL_OPERATION_TYPE",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "MUST_BE_CONTRACT",
    "type": "error"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "_code", "type": "uint256" }
    ],
    "name": "ONLY_LISTED_AGREEMENT",
    "type": "error"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bytes32",
        "name": "agreementType",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "code",
        "type": "address"
      }
    ],
    "name": "AgreementClassRegistered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bytes32",
        "name": "agreementType",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "code",
        "type": "address"
      }
    ],
    "name": "AgreementClassUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      }
    ],
    "name": "AppRegistered",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "bytes32",
        "name": "uuid",
        "type": "bytes32"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "codeAddress",
        "type": "address"
      }
    ],
    "name": "CodeUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract ISuperfluidGovernance",
        "name": "oldGov",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "contract ISuperfluidGovernance",
        "name": "newGov",
        "type": "address"
      }
    ],
    "name": "GovernanceReplaced",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "uint8",
        "name": "version",
        "type": "uint8"
      }
    ],
    "name": "Initialized",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "uint256",
        "name": "reason",
        "type": "uint256"
      }
    ],
    "name": "Jail",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": false,
        "internalType": "contract ISuperTokenFactory",
        "name": "newFactory",
        "type": "address"
      }
    ],
    "name": "SuperTokenFactoryUpdated",
    "type": "event"
  },
  {
    "anonymous": false,
    "inputs": [
      {
        "indexed": true,
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      },
      {
        "indexed": false,
        "internalType": "address",
        "name": "code",
        "type": "address"
      }
    ],
    "name": "SuperTokenLogicUpdated",
    "type": "event"
  },
  {
    "inputs": [],
    "name": "APP_WHITE_LISTING_ENABLED",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "CALLBACK_GAS_LIMIT",
    "outputs": [{ "internalType": "uint64", "name": "", "type": "uint64" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "NON_UPGRADABLE_DEPLOYMENT",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "bitmap", "type": "uint256" },
      { "internalType": "bytes32", "name": "agreementType", "type": "bytes32" }
    ],
    "name": "addToAgreementClassesBitmap",
    "outputs": [
      { "internalType": "uint256", "name": "newBitmap", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "targetApp",
        "type": "address"
      }
    ],
    "name": "allowCompositeApp",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes", "name": "ctx", "type": "bytes" },
      {
        "internalType": "int256",
        "name": "appCreditUsedDelta",
        "type": "int256"
      }
    ],
    "name": "appCallbackPop",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes", "name": "ctx", "type": "bytes" },
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      {
        "internalType": "uint256",
        "name": "appCreditGranted",
        "type": "uint256"
      },
      { "internalType": "int256", "name": "appCreditUsed", "type": "int256" },
      {
        "internalType": "contract ISuperfluidToken",
        "name": "appCreditToken",
        "type": "address"
      }
    ],
    "name": "appCallbackPush",
    "outputs": [{ "internalType": "bytes", "name": "appCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "uint32",
            "name": "operationType",
            "type": "uint32"
          },
          { "internalType": "address", "name": "target", "type": "address" },
          { "internalType": "bytes", "name": "data", "type": "bytes" }
        ],
        "internalType": "struct ISuperfluid.Operation[]",
        "name": "operations",
        "type": "tuple[]"
      }
    ],
    "name": "batchCall",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClass",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" },
      { "internalType": "bytes", "name": "userData", "type": "bytes" }
    ],
    "name": "callAgreement",
    "outputs": [
      { "internalType": "bytes", "name": "returnedData", "type": "bytes" }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClass",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" },
      { "internalType": "bytes", "name": "userData", "type": "bytes" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "callAgreementWithContext",
    "outputs": [
      { "internalType": "bytes", "name": "newCtx", "type": "bytes" },
      { "internalType": "bytes", "name": "returnedData", "type": "bytes" }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" }
    ],
    "name": "callAppAction",
    "outputs": [
      { "internalType": "bytes", "name": "returnedData", "type": "bytes" }
    ],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "callAppActionWithContext",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" },
      { "internalType": "bool", "name": "isTermination", "type": "bool" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "callAppAfterCallback",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "bytes", "name": "callData", "type": "bytes" },
      { "internalType": "bool", "name": "isTermination", "type": "bool" },
      { "internalType": "bytes", "name": "ctx", "type": "bytes" }
    ],
    "name": "callAppBeforeCallback",
    "outputs": [{ "internalType": "bytes", "name": "cbdata", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "castrate",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes", "name": "ctx", "type": "bytes" },
      {
        "internalType": "int256",
        "name": "appCreditUsedMore",
        "type": "int256"
      }
    ],
    "name": "ctxUseCredit",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "bytes", "name": "ctx", "type": "bytes" }],
    "name": "decodeCtx",
    "outputs": [
      {
        "components": [
          {
            "internalType": "uint8",
            "name": "appCallbackLevel",
            "type": "uint8"
          },
          { "internalType": "uint8", "name": "callType", "type": "uint8" },
          { "internalType": "uint256", "name": "timestamp", "type": "uint256" },
          { "internalType": "address", "name": "msgSender", "type": "address" },
          {
            "internalType": "bytes4",
            "name": "agreementSelector",
            "type": "bytes4"
          },
          { "internalType": "bytes", "name": "userData", "type": "bytes" },
          {
            "internalType": "uint256",
            "name": "appCreditGranted",
            "type": "uint256"
          },
          {
            "internalType": "uint256",
            "name": "appCreditWantedDeprecated",
            "type": "uint256"
          },
          {
            "internalType": "int256",
            "name": "appCreditUsed",
            "type": "int256"
          },
          {
            "internalType": "address",
            "name": "appAddress",
            "type": "address"
          },
          {
            "internalType": "contract ISuperfluidToken",
            "name": "appCreditToken",
            "type": "address"
          }
        ],
        "internalType": "struct ISuperfluid.Context",
        "name": "context",
        "type": "tuple"
      }
    ],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "components": [
          {
            "internalType": "uint32",
            "name": "operationType",
            "type": "uint32"
          },
          { "internalType": "address", "name": "target", "type": "address" },
          { "internalType": "bytes", "name": "data", "type": "bytes" }
        ],
        "internalType": "struct ISuperfluid.Operation[]",
        "name": "operations",
        "type": "tuple[]"
      }
    ],
    "name": "forwardBatchCall",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "agreementType", "type": "bytes32" }
    ],
    "name": "getAgreementClass",
    "outputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClass",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "appAddr",
        "type": "address"
      }
    ],
    "name": "getAppCallbackLevel",
    "outputs": [{ "internalType": "uint8", "name": "", "type": "uint8" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "contract ISuperApp", "name": "app", "type": "address" }
    ],
    "name": "getAppManifest",
    "outputs": [
      { "internalType": "bool", "name": "isSuperApp", "type": "bool" },
      { "internalType": "bool", "name": "isJailed", "type": "bool" },
      { "internalType": "uint256", "name": "noopMask", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getCodeAddress",
    "outputs": [
      { "internalType": "address", "name": "codeAddress", "type": "address" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getGovernance",
    "outputs": [
      {
        "internalType": "contract ISuperfluidGovernance",
        "name": "",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getNow",
    "outputs": [{ "internalType": "uint256", "name": "", "type": "uint256" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getSuperTokenFactory",
    "outputs": [
      {
        "internalType": "contract ISuperTokenFactory",
        "name": "factory",
        "type": "address"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "getSuperTokenFactoryLogic",
    "outputs": [
      { "internalType": "address", "name": "logic", "type": "address" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidGovernance",
        "name": "gov",
        "type": "address"
      }
    ],
    "name": "initialize",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClass",
        "type": "address"
      }
    ],
    "name": "isAgreementClassListed",
    "outputs": [{ "internalType": "bool", "name": "yes", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes32", "name": "agreementType", "type": "bytes32" }
    ],
    "name": "isAgreementTypeListed",
    "outputs": [{ "internalType": "bool", "name": "yes", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "contract ISuperApp", "name": "app", "type": "address" }
    ],
    "name": "isApp",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "contract ISuperApp", "name": "app", "type": "address" }
    ],
    "name": "isAppJailed",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      {
        "internalType": "contract ISuperApp",
        "name": "targetApp",
        "type": "address"
      }
    ],
    "name": "isCompositeAppAllowed",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [{ "internalType": "bytes", "name": "ctx", "type": "bytes" }],
    "name": "isCtxValid",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "forwarder", "type": "address" }
    ],
    "name": "isTrustedForwarder",
    "outputs": [{ "internalType": "bool", "name": "", "type": "bool" }],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "bytes", "name": "ctx", "type": "bytes" },
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "uint256", "name": "reason", "type": "uint256" }
    ],
    "name": "jailApp",
    "outputs": [{ "internalType": "bytes", "name": "newCtx", "type": "bytes" }],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "bitmap", "type": "uint256" }
    ],
    "name": "mapAgreementClasses",
    "outputs": [
      {
        "internalType": "contract ISuperAgreement[]",
        "name": "agreementClasses",
        "type": "address[]"
      }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "proxiableUUID",
    "outputs": [{ "internalType": "bytes32", "name": "", "type": "bytes32" }],
    "stateMutability": "pure",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClassLogic",
        "type": "address"
      }
    ],
    "name": "registerAgreementClass",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "configWord", "type": "uint256" }
    ],
    "name": "registerApp",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperApp",
        "name": "app",
        "type": "address"
      },
      { "internalType": "uint256", "name": "configWord", "type": "uint256" }
    ],
    "name": "registerAppByFactory",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "configWord", "type": "uint256" },
      { "internalType": "string", "name": "registrationKey", "type": "string" }
    ],
    "name": "registerAppWithKey",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "uint256", "name": "bitmap", "type": "uint256" },
      { "internalType": "bytes32", "name": "agreementType", "type": "bytes32" }
    ],
    "name": "removeFromAgreementClassesBitmap",
    "outputs": [
      { "internalType": "uint256", "name": "newBitmap", "type": "uint256" }
    ],
    "stateMutability": "view",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperfluidGovernance",
        "name": "newGov",
        "type": "address"
      }
    ],
    "name": "replaceGovernance",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperAgreement",
        "name": "agreementClassLogic",
        "type": "address"
      }
    ],
    "name": "updateAgreementClass",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      { "internalType": "address", "name": "newAddress", "type": "address" }
    ],
    "name": "updateCode",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperTokenFactory",
        "name": "newFactory",
        "type": "address"
      }
    ],
    "name": "updateSuperTokenFactory",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [
      {
        "internalType": "contract ISuperToken",
        "name": "token",
        "type": "address"
      }
    ],
    "name": "updateSuperTokenLogic",
    "outputs": [],
    "stateMutability": "nonpayable",
    "type": "function"
  },
  {
    "inputs": [],
    "name": "versionRecipient",
    "outputs": [{ "internalType": "string", "name": "", "type": "string" }],
    "stateMutability": "pure",
    "type": "function"
  }
]
'''



from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()

setup(
    name="aleph-superfluid",
    version="0.2.1",
    description="Fork of the Python SDK for the Superfluid Protocol",
    package_dir={"": "main"},
    packages=find_packages(where="main"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/aleph-im/superfluid.py",
    author="Godspower-Eze",
    author_email="Godspowereze260@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    install_requires=["web3 == 6.3.0", "python-decouple==3.8"],
    extras_require={
        "dev": ["twine>=4.0.2"]
    },
    python_requires=">=3"
)



from typing import Optional, TypedDict, Dict
from enum import Enum

from eth_typing import HexAddress, HexStr

from .utils import normalize_address, is_permissions_clean
from .errors import SFError


class GetFlowParams:

    def __init__(self, super_token: HexAddress, sender: HexAddress, receiver: HexAddress) -> None:
        """
            @param super_token - The token to be flowed
            @param sender - the sender of the flow
            @param receiver - the receiver of the flow
        """
        self.super_token = normalize_address(super_token)
        self.sender = normalize_address(sender)
        self.receiver = normalize_address(receiver)


class Web3FlowInfo(TypedDict):
    timestamp: int
    flowRate: int
    deposit: int
    owedDeposit: int


class GetAccountFlowInfoParams:

    def __init__(self, super_token: HexAddress, account: HexAddress) -> None:
        """
            @param super_token - The token to be flowed
            @param account - the account to get its info
        """
        self.super_token = normalize_address(super_token)
        self.account = normalize_address(account)


class GetFlowOperatorDataParams:

    def __init__(self, super_token: HexAddress, sender: HexAddress, flow_operator: HexAddress) -> None:
        """
            @param super_token - The token to be flowed
            @param sender - the sender of the flow
            @param flow_operator - the spender
        """
        self.super_token = normalize_address(super_token)
        self.sender = normalize_address(sender)
        self.flow_operator = normalize_address(flow_operator)


class Web3FlowOperatorData(TypedDict):
    flowOperatorId: HexStr
    permissions: int
    flowRateAllowance: int


class GetFlowOperatorDataParamsByID:

    def __init__(self, super_token: HexAddress, flow_operator_id: HexStr) -> None:
        """
            @param super_token - The token to be flowed
            @param flow_operator_id - the flow operator id
        """
        self.super_token = normalize_address(super_token)
        self.flow_operator_id = flow_operator_id


class ShouldUseCallAgreement:

    def __init__(self, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param should_use_call_agreement - whether or not to use the host contract
        """
        self.should_use_call_agreement = should_use_call_agreement


class UserData:

    def __init__(self, user_data: Optional[HexStr] = None) -> None:
        """
            @param user_data - Extra data provided
        """
        self.user_data = user_data


class ModifyFlowParams(ShouldUseCallAgreement, UserData):

    def __init__(self, receiver: HexAddress, super_token: HexAddress, flow_rate: Optional[int] = None, sender: Optional[HexAddress] = None, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param receiver - receiver of a flow
            @param super_token - The token to be flowed
            @param flow_rate(Optional) - flow rate for the flow
            @param sender(Optional) - sender of the flow
        """
        super().__init__(should_use_call_agreement)
        UserData.__init__(self, user_data)
        self.receiver = normalize_address(receiver)
        self.super_token = normalize_address(super_token)
        self.flow_rate = flow_rate
        self.sender = normalize_address(sender)


class CreateFlowParams(ModifyFlowParams):

    def __init__(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, flow_rate: int, user_data: Optional[HexStr] = None,  should_use_call_agreement: Optional[bool] = None) -> None:

        super().__init__(receiver,
                         super_token, sender=sender, flow_rate=flow_rate, user_data=user_data, should_use_call_agreement=should_use_call_agreement)


class UpdateFlowParams(CreateFlowParams):
    pass


class DeleteFlowParams(ModifyFlowParams):

    def __init__(self, sender: HexAddress, receiver: HexAddress, super_token: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        super().__init__(receiver,
                         super_token, sender=sender, user_data=user_data, should_use_call_agreement=should_use_call_agreement)


class CreateFlowByOperatorParams(CreateFlowParams):
    pass


class UpdateFlowByOperatorParams(CreateFlowByOperatorParams):
    pass


class SuperTokenFlowRateAllowanceParams(UserData):

    def __init__(self, flow_operator: HexAddress, flow_rate_allowance_delta: int, user_data: Optional[HexStr] = None) -> None:
        """
            @param flow_operator - The operator of the flow
            @param flow_rate_allowance_delta - The amount to increase the flow rate allowance by
        """
        super().__init__(user_data)
        self.flow_operator = normalize_address(flow_operator)
        self.flow_rate_allowance_delta = flow_rate_allowance_delta


class FlowRateAllowanceParams(SuperTokenFlowRateAllowanceParams):

    def __init__(self, super_token: HexAddress, flow_operator: HexAddress, flow_rate_allowance_delta: int, user_data: Optional[HexStr] = None) -> None:
        """ 
            @param flow_operator - The operator of the flow
            @param flow_rate_allowance_delta - The amount to increase the flow rate allowance by
            @param super_token - super token
        """
        super().__init__(flow_operator, flow_rate_allowance_delta, user_data)
        self.super_token = normalize_address(super_token)


class SuperTokenUpdateFlowOperatorPermissionsParams(ShouldUseCallAgreement, UserData):

    def __init__(self, flow_operator: HexAddress, permissions: int, flow_rate_allowance: int, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param flow_operator - The operator of the flow
            @param permissions - Number specifying the permission type
            @param flow_rate_allowance - Allowance to an operator
        """
        super().__init__(should_use_call_agreement)
        UserData.__init__(self, user_data)
        self.flow_operator = normalize_address(flow_operator)
        if is_permissions_clean(permissions):
            self.permissions = permissions
        else:
            raise SFError("UNCLEAN_PERMISSIONS",
                          "The desired permissions are unclean")
        if flow_rate_allowance < 0:
            raise SFError("NEGATIVE_FLOW_ALLOWANCE",
                          "No negative flow allowance allowed")
        else:
            self.flow_rate_allowance = flow_rate_allowance


class UpdateFlowOperatorPermissionsParams(SuperTokenUpdateFlowOperatorPermissionsParams):

    def __init__(self, super_token: HexAddress, flow_operator: HexAddress, permissions: int, flow_rate_allowance: int, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
            @param permissions - Number specifying the permission type
            @param flow_rate_allowance - Allowance to an operator
        """
        super().__init__(flow_operator, permissions,
                         flow_rate_allowance, user_data, should_use_call_agreement)
        self.super_token = normalize_address(super_token)


class SuperTokenFullControlParams(ShouldUseCallAgreement, UserData):

    def __init__(self, flow_operator: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param flow_operator - The operator of the flow
        """
        super().__init__(should_use_call_agreement)
        UserData.__init__(self, user_data)
        self.flow_operator = normalize_address(flow_operator)


class FullControlParams(SuperTokenFullControlParams):

    def __init__(self, super_token: HexAddress, flow_operator: HexAddress, user_data: Optional[HexStr] = None, should_use_call_agreement: Optional[bool] = None) -> None:
        """
            @param super_token - The token to be flowed
            @param flow_operator - The operator of the flow
        """
        super().__init__(flow_operator, user_data, should_use_call_agreement)
        self.super_token = super_token


class BatchOperationType(Enum):

    UNSUPPORTED = "UNSUPPORTED"  # 0
    ERC20_APPROVE = "ERC20_APPROVE"  # 1
    ERC20_TRANSFER_FROM = "ERC20_TRANSFER_FROM"  # 2
    ERC777_SEND = "ERC777_SEND"  # 3
    ERC20_INCREASE_ALLOWANCE = "ERC20_INCREASE_ALLOWANCE"  # 4
    ERC20_DECREASE_ALLOWANCE = "ERC20_DECREASE_ALLOWANCE"  # 5
    SUPERTOKEN_UPGRADE = "SUPERTOKEN_UPGRADE"  # 101
    SUPERTOKEN_DOWNGRADE = "SUPERTOKEN_DOWNGRADE"  # 102
    SUPERFLUID_CALL_AGREEMENT = "SUPERFLUID_CALL_AGREEMENT"  # 201
    CALL_APP_ACTION = "CALL_APP_ACTION"  # 202


class NETWORK:

    def __init__(self, network: Dict) -> None:
        self.CHAIN_ID = network["chainId"]
        self.RESOLVER = network["contractsV1"].get("resolver")
        self.HOST = network["contractsV1"].get("host")
        self.GOVERNANCE = network["contractsV1"].get("governance")
        self.CFA_V1 = network["contractsV1"].get("cfaV1")
        self.CFA_V1_FORWARDER = network["contractsV1"].get("cfaV1Forwarder")
        self.IDA_V1 = network["contractsV1"].get("idaV1")
        self.SUPER_TOKEN_FACTORY = network["contractsV1"].get(
            "superTokenFactory")
        self.SUPERFLUID_LOADER = network["contractsV1"].get("superfluidLoader")
        self.TOGA = network["contractsV1"].get("toga")
        self.BATCH_LIQUIDATOR = network["contractsV1"].get("batchLiquidator")
        self.FLOW_SCHEDULER = network["contractsV1"].get("flowScheduler")
        self.VESTING_SCHEDULER = network["contractsV1"].get("vestingScheduler")
        self.SUPER_SPREADER = network["contractsV1"].get("superSpreader")



class InvalidAddressError(Exception):
    pass


class InvalidChainId(Exception):
    pass


class SFError(Exception):

    def __init__(self, type: str, message: str, err: str = None) -> None:
        self.error_message = str(
            {"Type": type, "Message": message, "Error": err})
        super().__init__(self.error_message)
